@@ -22,7 +22,6 @@
 #define EGL_EGLEXT_PROTOTYPES
 #define GL_GLEXT_PROTOTYPES
 
-#include "gtk-compat.h"
 #include "spice-widget.h"
 #include "spice-widget-priv.h"
 #include "spice-gtk-session-priv.h"
@@ -31,6 +30,9 @@
 #endif
 
 #include <gdk/gdkx.h>
+#ifdef GDK_WINDOWING_WAYLAND
+#include <gdk/gdkwayland.h>
+#endif
 
 #define VERTS_ARRAY_SIZE (sizeof(GLfloat) * 4 * 4)
 #define TEX_ARRAY_SIZE (sizeof(GLfloat) * 4 * 2)
@@ -202,6 +204,14 @@ gboolean spice_egl_init(SpiceDisplay *display, GError **err)
     EGLNativeDisplayType dpy = 0;
     GdkDisplay *gdk_dpy = gdk_display_get_default();
 
+#ifdef GDK_WINDOWING_WAYLAND
+    if (GDK_IS_WAYLAND_DISPLAY(gdk_dpy)) {
+        d->egl.ctx = eglGetCurrentContext();
+        dpy = (EGLNativeDisplayType)gdk_wayland_display_get_wl_display(gdk_dpy);
+        d->egl.display = eglGetDisplay(dpy);
+        goto end;
+    }
+#endif
 #ifdef GDK_WINDOWING_X11
     if (GDK_IS_X11_DISPLAY(gdk_dpy)) {
         dpy = (EGLNativeDisplayType)gdk_x11_display_get_xdisplay(gdk_dpy);
@@ -258,7 +268,50 @@ gboolean spice_egl_init(SpiceDisplay *display, GError **err)
     eglMakeCurrent(d->egl.display, EGL_NO_SURFACE, EGL_NO_SURFACE,
                    d->egl.ctx);
 
-    return spice_egl_init_shaders(display, err);
+#ifdef GDK_WINDOWING_WAYLAND
+end:
+#endif
+
+    if (!spice_egl_init_shaders(display, err))
+        return FALSE;
+
+    d->egl.context_ready = TRUE;
+
+    if (spice_display_get_gl_scanout(SPICE_DISPLAY_CHANNEL(d->display)) != NULL) {
+        SPICE_DEBUG("scanout present during egl init, updating widget");
+        spice_display_widget_gl_scanout(display);
+        spice_display_widget_update_monitor_area(display);
+    }
+
+    return TRUE;
+}
+
+static gboolean
+gl_make_current(SpiceDisplay *display, GError **err)
+{
+    SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
+
+    if (GDK_IS_X11_DISPLAY(gdk_display_get_default())) {
+        EGLBoolean success = eglMakeCurrent(d->egl.display,
+                                            d->egl.surface,
+                                            d->egl.surface,
+                                            d->egl.ctx);
+        if (success != EGL_TRUE) {
+            g_set_error_literal(err, SPICE_CLIENT_ERROR,
+                                SPICE_CLIENT_ERROR_FAILED,
+                                "failed to activate context");
+            return FALSE;
+        }
+    }
+#if GTK_CHECK_VERSION(3,16,0)
+    else {
+        GtkWidget *area = gtk_stack_get_child_by_name(d->stack, "gl-area");
+
+        gtk_gl_area_make_current(GTK_GL_AREA(area));
+    }
+#endif
+
+    return TRUE;
 }
 
 static gboolean spice_widget_init_egl_win(SpiceDisplay *display, GdkWindow *win,
@@ -266,7 +319,6 @@ static gboolean spice_widget_init_egl_win(SpiceDisplay *display, GdkWindow *win,
 {
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
     EGLNativeWindowType native = 0;
-    EGLBoolean b;
 
     if (d->egl.surface)
         return TRUE;
@@ -293,15 +345,8 @@ static gboolean spice_widget_init_egl_win(SpiceDisplay *display, GdkWindow *win,
         return FALSE;
     }
 
-    b = eglMakeCurrent(d->egl.display,
-                       d->egl.surface,
-                       d->egl.surface,
-                       d->egl.ctx);
-    if (!b) {
-        g_set_error_literal(err, SPICE_CLIENT_ERROR, SPICE_CLIENT_ERROR_FAILED,
-                            "failed to activate context");
+    if (!gl_make_current(display, err))
         return FALSE;
-    }
 
     return TRUE;
 }
@@ -326,6 +371,9 @@ void spice_egl_unrealize_display(SpiceDisplay *display)
 
     SPICE_DEBUG("egl unrealize %p", d->egl.surface);
 
+    if (!gl_make_current(display, NULL))
+        return;
+
     if (d->egl.image != NULL) {
         eglDestroyImageKHR(d->egl.display, d->egl.image);
         d->egl.image = NULL;
@@ -341,10 +389,6 @@ void spice_egl_unrealize_display(SpiceDisplay *display)
         d->egl.tex_pointer_id = 0;
     }
 
-    if (d->egl.surface != EGL_NO_SURFACE) {
-        eglDestroySurface(d->egl.display, d->egl.surface);
-        d->egl.surface = EGL_NO_SURFACE;
-    }
     if (d->egl.vbuf_id) {
         glDeleteBuffers(1, &d->egl.vbuf_id);
         d->egl.vbuf_id = 0;
@@ -355,13 +399,26 @@ void spice_egl_unrealize_display(SpiceDisplay *display)
         d->egl.prog = 0;
     }
 
-    if (d->egl.ctx) {
-        eglDestroyContext(d->egl.display, d->egl.ctx);
-        d->egl.ctx = 0;
-    }
+    if (GDK_IS_X11_DISPLAY(gdk_display_get_default())) {
+        /* egl.surface && egl.ctx are only created on x11, see
+           spice_egl_init() */
 
-    eglMakeCurrent(d->egl.display, EGL_NO_SURFACE, EGL_NO_SURFACE,
-                   EGL_NO_CONTEXT);
+        if (d->egl.surface != EGL_NO_SURFACE) {
+            eglDestroySurface(d->egl.display, d->egl.surface);
+            d->egl.surface = EGL_NO_SURFACE;
+        }
+
+        if (d->egl.ctx) {
+            eglDestroyContext(d->egl.display, d->egl.ctx);
+            d->egl.ctx = 0;
+        }
+
+        eglMakeCurrent(d->egl.display, EGL_NO_SURFACE, EGL_NO_SURFACE,
+                       EGL_NO_CONTEXT);
+
+        /* do not call eglterminate() since egl may be used by
+         * somebody else code */
+    }
 }
 
 G_GNUC_INTERNAL
@@ -370,13 +427,16 @@ void spice_egl_resize_display(SpiceDisplay *display, int w, int h)
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
     int prog;
 
+    if (!gl_make_current(display, NULL))
+        return;
+
     glGetIntegerv(GL_CURRENT_PROGRAM, &prog);
 
     glUseProgram(d->egl.prog);
     apply_ortho(d->egl.mproj, 0, w, 0, h, -1, 1);
     glViewport(0, 0, w, h);
 
-    if (d->egl.image)
+    if (d->ready)
         spice_egl_update_display(display);
 
     glUseProgram(prog);
@@ -460,6 +520,8 @@ void spice_egl_cursor_set(SpiceDisplay *display)
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
     GdkPixbuf *image = d->mouse_pixbuf;
 
+    g_return_if_fail(d->egl.enabled);
+
     if (image == NULL)
         return;
 
@@ -486,8 +548,8 @@ void spice_egl_update_display(SpiceDisplay *display)
     gdouble tx, ty, tw, th;
     int prog;
 
-    g_return_if_fail(d->egl.image != NULL);
-    if (d->egl.surface == EGL_NO_SURFACE)
+    g_return_if_fail(d->ready);
+    if (!gl_make_current(display, NULL))
         return;
 
     spice_display_get_scaling(display, &s, &x, &y, &w, &h);
@@ -521,8 +583,14 @@ void spice_egl_update_display(SpiceDisplay *display)
     }
     SPICE_DEBUG("update %f +%d+%d %dx%d +%f+%f %fx%f", s, x, y, w, h,
                 tx, ty, tw, th);
-
     glBindTexture(GL_TEXTURE_2D, d->egl.tex_id);
+    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
+    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
+
+#if !GTK_CHECK_VERSION(3,16,0)
+    glEGLImageTargetTexture2DOES(GL_TEXTURE_2D, (GLeglImageOES)d->egl.image);
+#endif
+
     glDisable(GL_BLEND);
     glGetIntegerv(GL_CURRENT_PROGRAM, &prog);
     glUseProgram(d->egl.prog);
@@ -548,12 +616,14 @@ void spice_egl_update_display(SpiceDisplay *display)
                              0, 0, 1, 1);
     }
 
-    eglSwapBuffers(d->egl.display, d->egl.surface);
+    if (GDK_IS_X11_DISPLAY(gdk_display_get_default())) {
+        /* gtk+ does the swap with gtkglarea */
+        eglSwapBuffers(d->egl.display, d->egl.surface);
+    }
 
     glUseProgram(prog);
 }
 
-
 G_GNUC_INTERNAL
 gboolean spice_egl_update_scanout(SpiceDisplay *display,
                                   const SpiceGlScanout *scanout,
@@ -587,10 +657,11 @@ gboolean spice_egl_update_scanout(SpiceDisplay *display,
     attrs[10] = EGL_LINUX_DRM_FOURCC_EXT;
     attrs[11] = format;
     attrs[12] = EGL_NONE;
-    SPICE_DEBUG("fd:%d stride:%d y0:%d %dx%d format:0x%x (%c%c%c%c)",
+    SPICE_DEBUG("fd:%d stride:%u y0:%d %ux%u format:0x%x (%c%c%c%c)",
                 scanout->fd, scanout->stride, scanout->y0top,
                 scanout->width, scanout->height, format,
-                format & 0xff, (format >> 8) & 0xff, (format >> 16) & 0xff, format >> 24);
+                (int)format & 0xff, (int)(format >> 8) & 0xff,
+                (int)(format >> 16) & 0xff, (int)format >> 24);
 
     d->egl.image = eglCreateImageKHR(d->egl.display,
                                        EGL_NO_CONTEXT,
@@ -598,11 +669,15 @@ gboolean spice_egl_update_scanout(SpiceDisplay *display,
                                        (EGLClientBuffer)NULL,
                                        attrs);
 
-    glBindTexture(GL_TEXTURE_2D, d->egl.tex_id);
-    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
-    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
-    glEGLImageTargetTexture2DOES(GL_TEXTURE_2D, (GLeglImageOES)d->egl.image);
     d->egl.scanout = *scanout;
 
+#if GTK_CHECK_VERSION(3,16,0)
+    if (!gl_make_current(display, NULL))
+        return FALSE;
+
+    glBindTexture(GL_TEXTURE_2D, d->egl.tex_id);
+    glEGLImageTargetTexture2DOES(GL_TEXTURE_2D, (GLeglImageOES)d->egl.image);
+#endif
+
     return TRUE;
 }
